/* 
 * *********************************************************
 * Copyright (c) 2010 DHBW Mannheim - Tigers Mannheim
 * Project: tigers-centralSoftware
 * Date: 04.03.2010
 * Authors:
 * Bernhard Perun <bernhard.perun@googlemail.com>
 * *********************************************************
 */

package edu.moduli;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;

import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.SubnodeConfiguration;
import org.apache.commons.configuration.XMLConfiguration;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import edu.moduli.exceptions.DependencyException;
import edu.moduli.exceptions.InitModuleException;
import edu.moduli.exceptions.LoadModulesException;
import edu.moduli.exceptions.ModuleNotFoundException;
import edu.moduli.exceptions.StartModuleException;
import edu.moduli.listenerVariables.ModulesState;
import edu.moduli.listenerVariables.ModulesStateVariable;

/**
 * Main-class of moduli.
 * It contains the handeling of the modules.
 */
public class Moduli
{
	// --------------------------------------------------------------------------
	// --- instance variables ---------------------------------------------------
	// --------------------------------------------------------------------------

	// --- module path - indicates where the modules are located ---
	private String implsPath;
//	private String typesPath; // TODO: check type!

	// --- global configuration ---
	private SubnodeConfiguration globalConfiguration;
	
	// --- logger ---
	private Log log = LogFactory.getLog(this.getClass().getName());

	// --- moduleList ---
	ArrayList<AModule> moduleList = new ArrayList<AModule>();

	// --- module-state-variable ---
	private ModulesStateVariable modulesState = new ModulesStateVariable();
	
	// --------------------------------------------------------------------------
	// --- setter/getter --------------------------------------------------------
	// --------------------------------------------------------------------------

	/**
	 * Getter modulesState.
	 */
	public ModulesStateVariable getModulesState()
	{
		return modulesState;
	}

	/**
	 * Setter modulesState.
	 * Only to use if you know what you are doing ;).
	 */
	public void setModulesState(ModulesStateVariable modulesState)
	{
		this.modulesState = modulesState;
	}
	
	
	/**
	 * Getter global configuration
	 */
	public SubnodeConfiguration getGlobalConfiguration()
	{
		return globalConfiguration;
	}
	
	// --------------------------------------------------------------------------
	// --- public-method(s) -----------------------------------------------------
	// --------------------------------------------------------------------------

	/**
	 * Loads all available modules from configuration-file into modulesList.
	 * @param xmlFile (module-)configuration-file
	 * @throws LoadModulesException an error occurs... Can't continue.
	 */
	@SuppressWarnings("unchecked")
	public void loadModules(String xmlFile) throws LoadModulesException, DependencyException
	{
		// --- clear all moduleList ---
		moduleList.clear();
		
		// --- set state "NOT_LOADED" ---
		modulesState.set(ModulesState.NOT_LOADED);
		
		// --- fill it with new modules ---
		try 
		{
			XMLConfiguration config;

			config = new XMLConfiguration(xmlFile);
			
			// --- set moduli-folder ---
			implsPath = config.getString("moduliPath") + ".impls";
//			typesPath = config.getString("moduliPath") + ".types";
			
			// --- set globalConfiguration ---
			globalConfiguration = config.configurationAt("globalConfiguration");
			
			// --- load modules into modulesList ---
			for (int i=0; i<=config.getMaxIndex("module"); i++)
			{

				// --- create implementation- and properties-class ---
				Class<?> clazz = Class.forName(implsPath + "." + config.getString("module("+i+").implementation"));
				Class<?>[] propArgsClass = new Class[] {SubnodeConfiguration.class};

				// --- get properties from configuration and put it into a object[] ---
				SubnodeConfiguration moduleConfig = config.configurationAt("module("+i+").properties");
				Object[] propArgs = new Object[] {moduleConfig};

				// --- get constructor of implementation-class with subnodeConfiguration-parameter ---
				Constructor<?> clazzConstructor = clazz.getConstructor(propArgsClass);

				// --- create object (use constructor) ---
				AModule module = (AModule) createObject(clazzConstructor, propArgs);

				// --- set module config ---
				module.setSubnodeConfiguration(moduleConfig);
				
				// --- set id ---
				module.setId(config.getString("module("+i+")[@id]"));

				// --- set type ---
				module.setType(config.getString("module("+i+")[@type]"));

				// --- check if module is unique ---
				for(AModule m : moduleList)
				{
					if (m.getId().equals(module.getId()))
					{
						throw new LoadModulesException("module-id '" + module.getId() + "' isn't unique.");
					}
				}
				
				// --- check if module-type == config-type ---
				//TODO: really necessary???
				
				// --- set dependency-list ---
				ArrayList<String> depList = (ArrayList<String>) config.getList("module("+i+").dependency");
				module.setDependencies(depList);
				
				
				// --- put module into moduleList ---
				moduleList.add(module);

			}

		} 
		catch (ConfigurationException e) 
		{
			throw new LoadModulesException("Configuration contains errors: " + e.getMessage());
		} 
		catch (ClassNotFoundException e) 
		{
			throw new LoadModulesException("Class in configuration can't be found: " + e.getMessage());
		} 
		catch (SecurityException e) 
		{
			throw new LoadModulesException("Security issue at configuration : " + e.getMessage());
		} 
		catch (NoSuchMethodException e) 
		{
			throw new LoadModulesException("Can't find a constructor <init>(SubnodeConfiguration) of this class. Please add one. : " + e.getMessage());
		}
		catch (IllegalArgumentException e) 
		{
			throw new LoadModulesException("An argument isn't valid : " + e.getMessage());
		}

		// --- check dependencies ---
		checkDependencies();

		// --- set state "RESOLVED" ---
		modulesState.set(ModulesState.RESOLVED);
		
	}


	/**
	 * Starts all modules in modulesList.
	 */
	public void startModules() throws InitModuleException, StartModuleException
	{

		// --- init modules ---
		for(AModule m : moduleList)
		{
			m.initModule();
		}
		
		// --- wait a little bit ---
		try{ Thread.sleep(50); } catch (InterruptedException e) {}
		
		// --- start modules ---
		for(AModule m : moduleList)
		{
			m.startModule();
		}
		
		
		// --- set state "RESOLVED" ---
		modulesState.set(ModulesState.ACTIVE);
	}


	/**
	 * Stops all modules in modulesList.
	 */
	public void stopModules()
	{
		
		// --- stop modules ---
		for(AModule m : moduleList)
		{
			m.stopModule();
		}
		
		// --- wait a little bit ---
		try{ Thread.sleep(50); } catch (InterruptedException e) {}
		
		// --- deinit modules ---
		for(AModule m : moduleList)
		{
			m.deinitModule();
		}
		
		// --- set state "RESOLVED" ---
		modulesState.set(ModulesState.RESOLVED);
	}


	/**
	 * Returns a list with all loaded modules.
	 */
	public ArrayList<AModule> getModules() 
	{
		return moduleList;
	}
	
	/**
	 * Gets a module from current module-list.
	 * @param moduleId module-id-string
	 */
	public AModule getModule(String moduleId) throws ModuleNotFoundException
	{
		// --- search for the module ---
		for (AModule m : moduleList)
		{
			if ( m.getId().equals(moduleId) )
			{
				return m;
			}
		}

		// --- if nothing was found, throw a ModuleNotFoundException ---
		throw new ModuleNotFoundException("Module " + moduleId +" not found");
	}


	// --------------------------------------------------------------------------
	// --- private-method(s) ----------------------------------------------------
	// --------------------------------------------------------------------------

	/**
	 * Checks, if dependencies can be resolved.
	 * @throws DependencyException ... if at least one modules can't be resolved
	 */
	private void checkDependencies() throws DependencyException
	{

		// --- variable which indicates if dependencies are okay ---
		boolean depOk = false;

		// --- check if all dependencies can be resolved ---
		for(AModule m : moduleList)
		{

			for(String dependency : m.getDependencies())
			{
				// --- reset depOk ---
				depOk = false;

				for(AModule n : moduleList)
				{
					if (n.getId().equals(dependency))
					{
						// --- dep is okay ---
						depOk = true;
						break;
					}
				}

				// --- check if one dependencies isn't met ---
				if (depOk == false)
				{
					throw new DependencyException("Dependency '" + dependency +"' isn't met at module '" + m.getId() + "'");
				}

			}

		}
	}

	/**
	 * Creates an object from a constructor and its arguments.
	 */
	private Object createObject(Constructor<?> constructor, Object[] arguments) 
	{
		Object object = null;

		try {
			object = constructor.newInstance(arguments);
			return object;
		} catch (InstantiationException e) {
			log.error(e.getMessage());
		} catch (IllegalAccessException e) {
			log.error(e.getMessage());
		} catch (IllegalArgumentException e) {
			log.error(e.getMessage());
		} catch (InvocationTargetException e) {
			log.error(e.getMessage());
		}
		return object;
	}
	
//	/**
//	 * Gets all supertype of a class.
//	 * NOT USED AT THE MOMENT
//	 * @param o Object
//	 * @return ArrayList with Superclasses
//	 */
//	private ArrayList<String> getSuperclasses(Object o) {
//		Class<?> subclass = o.getClass();
//		Class<?> superclass = subclass.getSuperclass();
//		ArrayList<String> list = new ArrayList<String>();
//		while (superclass != null) {
//			String className = superclass.getName();
//			list.add(className);
//			subclass = superclass;
//			superclass = subclass.getSuperclass();
//		}
//		return list;
//	}
}
